package sdanderson.androidcharts.LineChart;

import android.content.Context;
import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.drawable.BitmapDrawable;
import android.util.AttributeSet;
import android.util.TypedValue;
import android.view.Gravity;
import android.view.LayoutInflater;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.widget.Adapter;
import android.widget.LinearLayout;
import android.widget.ListView;
import android.widget.PopupWindow;
import android.widget.RelativeLayout;
import android.widget.TextView;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;

import sdanderson.androidcharts.Common.Coordinate;
import sdanderson.androidcharts.Common.NumberUtilities;
import sdanderson.androidcharts.R;

/**
 * Created by Shane Anderson on 9/30/2015.
 *
 */
public class LineChart extends RelativeLayout {

    private TextView mTextViewTitle;
    private TextView mTextViewYAxisLabel;
    private TextView mTextViewXAxisLabel;
    private TextView mTextViewYAxisLargestValue;
    private TextView mTextViewYAxisSmallestValue;
    private TextView mTextViewXAxisLargestValue;
    private TextView mTextViewXAxisSmallestValue;
    private boolean mUseNumberAbbreviationSuffixXValueChart = false;
    private boolean mUseNumberAbbreviationSuffixYValueChart = false;
    private boolean mUseNumberAbbreviationSuffixXValuePopup = false;
    private boolean mUseNumberAbbreviationSuffixYValuePopup = false;
    private DecimalFormat mDecimalFormatChartX = new DecimalFormat("#.#");
    private DecimalFormat mDecimalFormatChartY = new DecimalFormat("#.#");
    private DecimalFormat mDecimalFormatPopupX = new DecimalFormat("#.#");
    private DecimalFormat mDecimalFormatPopupY = new DecimalFormat("#.#");
    private String mNumberSuffixes = "kmbt";
    private int mNumberGroupSplit = 3;

    PopupWindow mLingeringPopupWindow = null;

    private LineChartInnerView mLineChartInnerView;

    private OnChartTouchedListener mOnChartTouchedListener = null;

    public LineChart (Context pContext) {
        super(pContext);
        init(pContext);
    }

    public LineChart(Context pContext, AttributeSet pAttrs) {
        super(pContext, pAttrs);
        init(pContext);
    }

    private void init(Context pContext) {
        Paint paint;
        setBackgroundColor(Color.TRANSPARENT);
        paint = new Paint(Paint.ANTI_ALIAS_FLAG);
        paint.setStyle(Paint.Style.FILL);

        View.inflate(pContext, R.layout.linechartrelative, this);
        setLayoutParams(new LinearLayout.LayoutParams(
                LayoutParams.MATCH_PARENT, LayoutParams.MATCH_PARENT));
        mLineChartInnerView = (LineChartInnerView) findViewById(R.id.linechart_chart);
        mTextViewTitle = (TextView) findViewById(R.id.linechart_title);
        mTextViewTitle.setText(" ");
        mTextViewXAxisLabel = (TextView) findViewById(R.id.linechart_xaxis_label);
        mTextViewXAxisLabel.setText(" ");
        mTextViewYAxisLabel = (TextView) findViewById(R.id.linechart_yaxis_label);
        mTextViewYAxisLabel.setText(" ");
        mTextViewXAxisLargestValue = (TextView) findViewById(R.id.linechart_xaxishighestvalue);
        mTextViewXAxisSmallestValue = (TextView) findViewById(R.id.linechart_xaxislowestvalue);
        mTextViewYAxisLargestValue = (TextView) findViewById(R.id.linechart_yaxishighestvalue);
        mTextViewYAxisSmallestValue = (TextView) findViewById(R.id.linechart_yaxislowestvalue);

        mLineChartInnerView.setOnTouchListener(new OnTouchListener() {
            float mX = 0, mY = 0;

            @Override
            public boolean onTouch(View view, MotionEvent motionEvent) {
                if (motionEvent.getAction() == MotionEvent.ACTION_DOWN) {
                    mX = motionEvent.getX();
                    mY = motionEvent.getY();
                }
                if (motionEvent.getEventTime() - motionEvent.getDownTime() < 500 &&
                        motionEvent.getAction() == MotionEvent.ACTION_UP) {
                    if (mOnChartTouchedListener != null) {
                        mOnChartTouchedListener.onChartTouched(
                                mLineChartInnerView.getTappedLineCoordIndexes(mX, mY),
                                LineChart.this, mX, mY);
                    } else {
                        List<Coordinate<Integer, Integer>> touchedPoints =
                                mLineChartInnerView.getTappedLineCoordIndexes(mX, mY);
                        if (touchedPoints.size() == 1)
                            defaultTouchResponseSingle(touchedPoints.get(0).getX(),
                                    touchedPoints.get(0).getY(), mX, mY);
                        else if (touchedPoints.size() > 1)
                            defaultTouchResponseMultiple(touchedPoints, mX, mY);
                    }
                }
                return true;
            }
        });
    }

    @Override
    protected void onLayout(boolean pChanged, int pLeft, int pTop, int pRight, int pBottom) {
        super.onLayout(pChanged, pLeft, pTop, pRight, pBottom);
    }

    @Override
    protected void onSizeChanged(int pWidth, int pHeight, int pOldWidth, int pOldHeight) {
        super.onSizeChanged(pWidth, pHeight, pOldWidth, pOldHeight);
    }

    @Override
    protected void onDetachedFromWindow() {
        super.onDetachedFromWindow();
        if (mLingeringPopupWindow != null && mLingeringPopupWindow.isShowing()) {
            mLingeringPopupWindow.dismiss();
        }
    }

    private void defaultTouchResponseSingle(int pLineIndex, int pCoordIndex, float pTouchedX,
                                            float pTouchedY) {
        Resources res = getResources();
        String lineLabel = mLineChartInnerView.getLineLabel(pLineIndex);
        Coordinate<Float, Float> coordinate =
                mLineChartInnerView.getCoordinateAtIndexes(pLineIndex, pCoordIndex);
        int lineColor = mLineChartInnerView.getLineColor(pLineIndex);

        LayoutInflater inflater = (LayoutInflater)
                getContext().getSystemService(Context.LAYOUT_INFLATER_SERVICE);
        View v = inflater.inflate(R.layout.linechart_popupwindow, (ViewGroup) getParent(), false);
        TextView tv = (TextView) v.findViewById(R.id.linechart_popupwindow_linetitletext);
        tv.setText(lineLabel);
        tv = (TextView) v.findViewById(R.id.linechart_popupwindow_xaxisvalue);
        if (mUseNumberAbbreviationSuffixXValuePopup)
            tv.setText(String.format(res.getString(R.string.colonspace),
                                    mTextViewXAxisLabel.getText(),
                                    NumberUtilities.formatValueWithSuffix(coordinate.getX(),
                                        mDecimalFormatPopupX, mNumberSuffixes, mNumberGroupSplit)));
        else {
            tv.setText(String.format(res.getString(R.string.colonspace),
                                    mTextViewXAxisLabel.getText(),
                                    mDecimalFormatPopupX.format(coordinate.getX())));
        }
        tv = (TextView) v.findViewById(R.id.linechart_popupwindow_yaxisvalue);
        if (mUseNumberAbbreviationSuffixYValuePopup)
            tv.setText(String.format(res.getString(R.string.colonspace),
                                    mTextViewYAxisLabel.getText(),
                                    NumberUtilities.formatValueWithSuffix(coordinate.getY(),
                                        mDecimalFormatPopupY, mNumberSuffixes, mNumberGroupSplit)));
        else {
            tv.setText(String.format(res.getString(R.string.colonspace),
                                    mTextViewYAxisLabel.getText(),
                                    mDecimalFormatPopupY.format(coordinate.getY())));
        }
        View colorkey = v.findViewById(R.id.linechart_popupwindow_colorkey);
        colorkey.setBackgroundColor(lineColor);
        final PopupWindow pw = new PopupWindow(v, WindowManager.LayoutParams.WRAP_CONTENT,
                                                    WindowManager.LayoutParams.WRAP_CONTENT);
        mLingeringPopupWindow = pw;
        // Need to set a background drawable to some random bitmap to make it touchable (bug?)
        pw.setBackgroundDrawable(new BitmapDrawable(getContext().getResources(),
                Bitmap.createBitmap(1, 1, Bitmap.Config.ARGB_8888)));
        pw.setOutsideTouchable(true);
        pw.setTouchInterceptor(new View.OnTouchListener() {
            @Override
            public boolean onTouch(View view, MotionEvent motionEvent) {
                pw.dismiss();
                return true;
            }
        });
        View parentView = (View) getParent();
        int[] location = new int[2];
        LineChart.this.getLocationOnScreen(location);
        Resources r = getResources();
        float pxY   = TypedValue.applyDimension(TypedValue.COMPLEX_UNIT_DIP, 80,
                r.getDisplayMetrics());
        pw.showAtLocation(parentView, Gravity.NO_GRAVITY,
                location[0] + (int) pTouchedX, location[1] + (int) pTouchedY - (int) (pxY));
    }

    private int getWidestView(Context pContext, Adapter pAdapter) {
        int maxWidth = 0;
        View view = null;
        RelativeLayout fakeParent = new RelativeLayout(pContext);
        for (int i=0, count=pAdapter.getCount(); i<count; i++) {
            view = pAdapter.getView(i, view, fakeParent);
            view.measure(View.MeasureSpec.UNSPECIFIED, View.MeasureSpec.UNSPECIFIED);
            int width = view.getMeasuredWidth();
            if (width > maxWidth) {
                maxWidth = width;
            }
        }
        return maxWidth;
    }

    private void defaultTouchResponseMultiple(List<Coordinate<Integer, Integer>> pTouchedPoints,
                                              float pTouchedX, float pTouchedY) {
        List<String> labels = new ArrayList<>();
        List<Integer> colors = new ArrayList<>();
        List<Coordinate<Float, Float>> coordinates = new ArrayList<>();

        for (Coordinate<Integer, Integer> listcoordPair:pTouchedPoints) {
            labels.add(mLineChartInnerView.getLineLabel(listcoordPair.getX()));
            colors.add(mLineChartInnerView.getLineColor(listcoordPair.getX()));
            coordinates.add(mLineChartInnerView.getCoordinateAtIndexes(listcoordPair.getX(),
                    listcoordPair.getY()));
        }

        LayoutInflater inflater = (LayoutInflater)
                getContext().getSystemService(Context.LAYOUT_INFLATER_SERVICE);
        View v = inflater.inflate(R.layout.linechart_popupwindow_multiplepoints,
                                (ViewGroup) getParent(), false);
        ListView listview =
                (ListView)v.findViewById(R.id.linechart_popupwindow_multiplepoints_listview);
        MultipleDataPointsAdapter adapter = new MultipleDataPointsAdapter(this.getContext(),
                coordinates, labels, colors, mTextViewXAxisLabel.getText().toString(),
                mTextViewYAxisLabel.getText().toString(), mUseNumberAbbreviationSuffixXValuePopup,
                mUseNumberAbbreviationSuffixYValuePopup, mDecimalFormatPopupX, mDecimalFormatPopupY,
                mNumberSuffixes, mNumberGroupSplit);
        listview.setAdapter(adapter);
        listview.getLayoutParams().width = getWidestView(this.getContext(), adapter);

        final PopupWindow pw = new PopupWindow(v, WindowManager.LayoutParams.WRAP_CONTENT,
                                                    WindowManager.LayoutParams.WRAP_CONTENT);
        mLingeringPopupWindow = pw;
        // Need to set a background drawable to some random bitmap to make it touchable (bug?)
        pw.setBackgroundDrawable(new BitmapDrawable(getContext().getResources(),
                Bitmap.createBitmap(1, 1, Bitmap.Config.ARGB_8888)));
        pw.setOutsideTouchable(true);
        pw.setTouchInterceptor(new View.OnTouchListener() {
            @Override
            public boolean onTouch(View view, MotionEvent motionEvent) {
                pw.dismiss();
                return true;
            }
        });
        View parentView = (View) getParent();
        int[] location = new int[2];
        LineChart.this.getLocationOnScreen(location);
        Resources r = getResources();
        float pxY   = TypedValue.applyDimension(TypedValue.COMPLEX_UNIT_DIP, 80,
                r.getDisplayMetrics());
        pw.showAtLocation(parentView, Gravity.NO_GRAVITY,
                location[0]+(int)pTouchedX, location[1]+(int)pTouchedY-(int)(pxY));
    }

    private void refreshAxisValueText() {
        if (mLineChartInnerView.areThereLines()) {
            if (mUseNumberAbbreviationSuffixXValueChart) {
                mTextViewXAxisSmallestValue.setText(
                        NumberUtilities.formatValueWithSuffix(
                                mLineChartInnerView.getXAxisSmallestValue(), mDecimalFormatChartX,
                                mNumberSuffixes, mNumberGroupSplit));
                mTextViewXAxisLargestValue.setText(
                        NumberUtilities.formatValueWithSuffix(
                                mLineChartInnerView.getXAxisLargestValue(), mDecimalFormatChartX,
                                mNumberSuffixes, mNumberGroupSplit));
            } else {
                mTextViewXAxisSmallestValue.setText(mDecimalFormatChartX.format(
                                                mLineChartInnerView.getXAxisSmallestValue()));
                mTextViewXAxisLargestValue.setText(mDecimalFormatChartX.format(
                                                mLineChartInnerView.getXAxisLargestValue()));
            }

            if (mUseNumberAbbreviationSuffixYValueChart) {
                mTextViewYAxisSmallestValue.setText(NumberUtilities.formatValueWithSuffix(
                        mLineChartInnerView.getYAxisSmallestValue(), mDecimalFormatChartY,
                        mNumberSuffixes, mNumberGroupSplit));
                mTextViewYAxisLargestValue.setText(NumberUtilities.formatValueWithSuffix(
                        mLineChartInnerView.getYAxisLargestValue(), mDecimalFormatChartY,
                        mNumberSuffixes, mNumberGroupSplit));
            } else {
                mTextViewYAxisSmallestValue.setText(mDecimalFormatChartY.format(
                        mLineChartInnerView.getYAxisSmallestValue()));
                mTextViewYAxisLargestValue.setText(mDecimalFormatChartY.format(
                        mLineChartInnerView.getYAxisLargestValue()));
            }
        }
    }

    // PUBLICLY AVAILABLE INTERACTION METHODS

    @SuppressWarnings("unused")
    public void addLine(String pLabel, List<Coordinate<Float, Float>> pCoordinates, int pColor) {
        mLineChartInnerView.addLine(pLabel, pCoordinates, pColor);
        refreshAxisValueText();
    }

    @SuppressWarnings("unused")
    public void setChartTitle(String pTitle) { mTextViewTitle.setText(pTitle); }

    @SuppressWarnings("unused")
    public void setChartTitleColor(int pColor) { mTextViewTitle.setTextColor(pColor); }

    @SuppressWarnings("unused")
    public void setChartBackgroundColor(int pColor) {
        mLineChartInnerView.setChartBackgroundColor(pColor);
    }

    @SuppressWarnings("unused")
    public void setGridlineWidth(float pGridlineWidth) {
        mLineChartInnerView.setGridlineWidth(pGridlineWidth);
    }

    // This is to set the number of gridlines on both sides of the Y=0 line
    @SuppressWarnings("unused")
    public void setNumOfGridlines(int pNumOfGridlines) {
        mLineChartInnerView.setNumOfGridlines(pNumOfGridlines);
    }

    @SuppressWarnings("unused")
    public void setChartGridlineColor(int pColor) { mLineChartInnerView.setGridlineColor(pColor); }

    @SuppressWarnings("unused")
    public void setZeroLineWidth(float pWidth) { mLineChartInnerView.setZeroLineWidth(pWidth); }

    @SuppressWarnings("unused")
    public void setZeroLineColor(int pColor) { mLineChartInnerView.setZeroLineColor(pColor); }

    @SuppressWarnings("unused")
    public void setRoundOutChartYMinMax(boolean pRound) { setRoundOutChartYMinMax(pRound, 1); }

    @SuppressWarnings("unused")
    public void setRoundOutChartYMinMax(boolean pRound, int pHighestDigits) {
        mLineChartInnerView.setRoundOutChartYMinMax(pRound, pHighestDigits);
    }

    @SuppressWarnings("unused")
    public void setRoundOutChartXMinMax(boolean pRound) { setRoundOutChartYMinMax(pRound, 1); }

    @SuppressWarnings("unused")
    public void setRoundOutChartXMinMax(boolean pRound, int pHighestDigits) {
        mLineChartInnerView.setRoundOutChartXMinMax(pRound, pHighestDigits);
    }

    @SuppressWarnings("unused")
    public void setBorderWidth(float pWidth) { mLineChartInnerView.setBorderWidth(pWidth); }

    @SuppressWarnings("unused")
    public void setBorderColor(int pColor) { mLineChartInnerView.setBorderColor(pColor); }

    @SuppressWarnings("unused")
    public void setXLabelColor(int pColor) { mTextViewXAxisLabel.setTextColor(pColor); }

    @SuppressWarnings("unused")
    public void setYLabelColor(int pColor) {
        mTextViewYAxisLabel.setTextColor(pColor);
    }

    @SuppressWarnings("unused")
    public void setXValueLabelsColor(int pColor) {
        mTextViewXAxisLargestValue.setTextColor(pColor);
        mTextViewXAxisSmallestValue.setTextColor(pColor);
    }

    @SuppressWarnings("unused")
    public void setYValueLabelsColor(int pColor) {
        mTextViewYAxisLargestValue.setTextColor(pColor);
        mTextViewYAxisSmallestValue.setTextColor(pColor);
    }

    @SuppressWarnings("unused")
    public void setXLabel(String pLabel) { mTextViewXAxisLabel.setText(pLabel); }

    @SuppressWarnings("unused")
    public void setYLabel(String pLabel) {
        mTextViewYAxisLabel.setText(pLabel);
    }

    @SuppressWarnings("unused")
    public void setDecimalFormatChartX(DecimalFormat pDecimalFormat) {
        mDecimalFormatChartX = pDecimalFormat;
        refreshAxisValueText();
    }

    @SuppressWarnings("unused")
    public void setDecimalFormatChartY(DecimalFormat pDecimalFormat) {
        mDecimalFormatChartY = pDecimalFormat;
        refreshAxisValueText();
    }

    @SuppressWarnings("unused")
    public void setDecimalFormatPopupX(DecimalFormat pDecimalFormat) {
        mDecimalFormatPopupX = pDecimalFormat;
    }

    @SuppressWarnings("unused")
    public void setDecimalFormatPopupY(DecimalFormat pDecimalFormat) {
        mDecimalFormatPopupY = pDecimalFormat;
    }

    @SuppressWarnings("unused")
    public void setUseNumberAbbreviationSuffixXValueChart(boolean pUseSuffix) {
        mUseNumberAbbreviationSuffixXValueChart = pUseSuffix;
        refreshAxisValueText();
    }

    @SuppressWarnings("unused")
    public void setUseNumberAbbreviationSuffixYValueChart(boolean pUseSuffix) {
        mUseNumberAbbreviationSuffixYValueChart = pUseSuffix;
        refreshAxisValueText();
    }

    @SuppressWarnings("unused")
    public void setUseNumberAbbreviationSuffixXValuePopup(boolean pUseSuffix) {
        mUseNumberAbbreviationSuffixXValuePopup = pUseSuffix;
    }

    @SuppressWarnings("unused")
    public void setUseNumberAbbreviationSuffixYValuePopup(boolean pUseSuffix) {
        mUseNumberAbbreviationSuffixYValuePopup = pUseSuffix;
    }

    @SuppressWarnings("unused")
    public void setNumberSuffixes(String pSuffixes) {
        mNumberSuffixes = pSuffixes;
        refreshAxisValueText();
    }

    @SuppressWarnings("unused")
    public void setNumberGroupSplit(int pGroupSplit) {
        mNumberGroupSplit = pGroupSplit;
        refreshAxisValueText();
    }

    @SuppressWarnings("unused")
    public void setUseAnimation(boolean pUseAnimation) {
        mLineChartInnerView.setUseAnimation(pUseAnimation);
    }

    @SuppressWarnings("unused")
    public void setAnimationTime(float pAnimationTime) {
        mLineChartInnerView.setAnimationTime(pAnimationTime);
    }

    @SuppressWarnings("unused")
    public void reanimate() { mLineChartInnerView.reanimate(); }

    @SuppressWarnings("unused")
    public void setOnChartTouchedListener(OnChartTouchedListener pOnChartTouchedListener) {
        mOnChartTouchedListener = pOnChartTouchedListener;
    }

    @SuppressWarnings("unused")
    public interface OnChartTouchedListener {
        void onChartTouched(List<Coordinate<Integer, Integer>> pTouchedPoints, LineChart pLineChart,
                            float pTouchedX, float pTouchedY);
    }
}
